/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */

package fr.amap.lidar.amapvox.engine3d.views;

import fr.amap.lidar.amapvox.engine3d.util.InputStatus;
import com.jogamp.newt.event.InputEvent;
import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL3bc;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLDebugMessage;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.util.GLBuffers;
import fr.amap.lidar.amapvox.engine3d.entities.base.Scene;
import fr.amap.lidar.amapvox.engine3d.resources.postprocesses.base.FrameBuffer;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.imageio.ImageIO;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * OpenGL loop manager of this project.
 */
@ThreadSafe
public class Listener3D implements GLEventListener {

    /**
     * The current scene.
     */
    @GuardedBy(value = "this")
    private Scene scene;
    
    /**
     * Start time of this application.
     */
    @GuardedBy(value = "this")
    private final long startTime;
    
    /**
     * The last time os each updaing.
     */
    @GuardedBy(value = "this")
    private long lastTime;
    
    /**
     * If the scene has changed.
     */
    @GuardedBy(value = "this")
    private boolean newScene;
    
    /**
     * If the scene has changed.
     */
    @GuardedBy(value = "this")
    private boolean takeScreenshot;
    
    /**
     * If the scene has changed.
     */
    @GuardedBy(value = "this")
    private BufferedImage screenshot;
    
    /**
     * Screen dimensions.
     */
    @GuardedBy(value = "this")
    private int width, height;
    
    /**
     * The current OpenGL context.
     */
    @GuardedBy(value = "this")
    private GL gl;

    /**
     * Create a new GL event listener.
     */
    public Listener3D() {
        this.scene = null;
        this.startTime = System.currentTimeMillis();
        this.newScene = false;
        this.gl = null;
    }
    
    /**
     * Set the current scene.
     * @param scene 
     */
    public final synchronized void setScene(final Scene scene) {
        if (scene != null && gl != null) {
            scene.destroy(gl);
        }
        
        this.scene = scene;
        this.newScene = true;
    }

    public final synchronized void takeScreenshot() {
        this.takeScreenshot = true;
    }
    
    /**
     * Get the current scene.
     * @return 
     */
    public final synchronized Scene getScene() {
        return scene;
    }

    public final synchronized GL getGL() {
        return gl;
    }

    public final BufferedImage getScreenshot() {
        return screenshot;
    }
    
    @Override
    public final void init(final GLAutoDrawable drawable) {
        this.gl = drawable.getGL();
                
        /* Initialize the visualization context */
        System.setProperty("jogl.disable.openglcore", "true");
        
        System.out.println("Chosen GLCapabilities: " + drawable.getChosenGLCapabilities());
        System.out.println("INIT GL IS: " + gl.getClass().getName());
        System.out.println("GL_VENDOR: " + gl.glGetString(GL3bc.GL_VENDOR));
        System.out.println("GL_RENDERER: " + gl.glGetString(GL3bc.GL_RENDERER));
        System.out.println("GL_VERSION: " + gl.glGetString(GL3bc.GL_VERSION));
                
        drawable.getContext().addGLDebugListener((GLDebugMessage event) -> {
            if (event.getDbgSeverity() == GL3bc.GL_DEBUG_SEVERITY_HIGH) {
                System.out.println("GL Debug: " + event.getDbgMsg());
            }
        });
        
        if (scene != null) {
            scene.init(gl);
        }
    }

    @Override
    public final void display(final GLAutoDrawable drawable) {
        this.gl = drawable.getGL();
        
        if (newScene) {
            scene.init(gl);
            scene.rescale(gl, width, height);
            newScene = false;
        }
        
        if (scene != null) {
            long time = startTime - System.currentTimeMillis();
            scene.update(gl, (lastTime - time) * 0.01f);
            lastTime = time;
            scene.display(gl);
        } else {
            gl.glClearColor(0.92f, 0.93f, 0.95f, 1.0f);
            gl.glClear(GL3bc.GL_COLOR_BUFFER_BIT);
        }
        
        if (takeScreenshot) {
            screenshot = new BufferedImage(width, 
                height, BufferedImage.TYPE_INT_RGB);
            final Graphics graphics = screenshot.getGraphics();

            final ByteBuffer buffer = GLBuffers.newDirectByteBuffer(
                    width * height * 4);

            gl.getGL4bc().glReadBuffer(GL.GL_BACK);
            gl.getGL4bc().glReadPixels(0, 0, width, height, 
                    GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, buffer);

            for (int h = 0; h < height; h++) {
                for (int w = 0; w < width; w++) {
                    graphics.setColor(new Color((buffer.get() & 0xff), (buffer.get() & 0xff),
                            (buffer.get() & 0xff)));
                    buffer.get();
                    graphics.drawRect(w, height - h, 1, 1);
                }
            }
            
            takeScreenshot = false;
        }
    }
    
    public final synchronized void inputting(final InputEvent event, final InputStatus status) {
        if (scene != null) {
            scene.onInputting(event, status);
        }
    }

    @Override
    public final void reshape(final GLAutoDrawable drawable, final int x, 
            final int y, final int width, final int height) {
        this.gl = drawable.getGL().getGL3bc();
        this.width = width;
        this.height = height;
        
        if (scene != null) {
            scene.rescale(gl, width, height);
        }
    }
    
    @Override
    public final void dispose(final GLAutoDrawable drawable) {
        gl = drawable.getGL().getGL3bc();
        
        if (scene != null) {
            scene.destroy(gl);
        }
    }

}
